package org.tasks.gtasks;

import static com.natpryce.makeiteasy.MakeItEasy.with;
import static com.todoroo.astrid.data.Task.HIDE_UNTIL_DUE;
import static com.todoroo.astrid.data.Task.HIDE_UNTIL_DUE_TIME;
import static junit.framework.Assert.assertEquals;
import static junit.framework.Assert.assertNull;
import static org.tasks.gtasks.GoogleTaskSynchronizer.getTruncatedValue;
import static org.tasks.gtasks.GoogleTaskSynchronizer.mergeDates;
import static org.tasks.gtasks.GoogleTaskSynchronizer.truncate;
import static org.tasks.makers.TaskMaker.DUE_DATE;
import static org.tasks.makers.TaskMaker.DUE_TIME;
import static org.tasks.makers.TaskMaker.HIDE_TYPE;
import static org.tasks.makers.TaskMaker.newTask;

import androidx.test.ext.junit.runners.AndroidJUnit4;
import com.todoroo.astrid.data.Task;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.tasks.time.DateTime;

@RunWith(AndroidJUnit4.class)
public class GoogleTaskSynchronizerTest {

  @Test
  public void testMergeDate() {
    Task local = newTask(with(DUE_DATE, new DateTime(2016, 3, 12)));

    mergeDates(newTask(with(DUE_DATE, new DateTime(2016, 3, 11))).getDueDate(), local);

    assertEquals(new DateTime(2016, 3, 11, 12, 0).getMillis(), local.getDueDate().longValue());
  }

  @Test
  public void testMergeTime() {
    Task local = newTask(with(DUE_TIME, new DateTime(2016, 3, 11, 13, 30)));

    mergeDates(newTask(with(DUE_DATE, new DateTime(2016, 3, 11))).getDueDate(), local);

    assertEquals(new DateTime(2016, 3, 11, 13, 30, 1).getMillis(), local.getDueDate().longValue());
  }

  @Test
  public void testDueDateAdjustHideBackwards() {
    Task local =
        newTask(with(DUE_DATE, new DateTime(2016, 3, 12)), with(HIDE_TYPE, HIDE_UNTIL_DUE));

    mergeDates(newTask(with(DUE_DATE, new DateTime(2016, 3, 11))).getDueDate(), local);

    assertEquals(new DateTime(2016, 3, 11).getMillis(), local.getHideUntil().longValue());
  }

  @Test
  public void testDueDateAdjustHideForwards() {
    Task local =
        newTask(with(DUE_DATE, new DateTime(2016, 3, 12)), with(HIDE_TYPE, HIDE_UNTIL_DUE));

    mergeDates(newTask(with(DUE_DATE, new DateTime(2016, 3, 14))).getDueDate(), local);

    assertEquals(new DateTime(2016, 3, 14).getMillis(), local.getHideUntil().longValue());
  }

  @Test
  public void testDueTimeAdjustHideBackwards() {
    Task local =
        newTask(
            with(DUE_TIME, new DateTime(2016, 3, 12, 13, 30)),
            with(HIDE_TYPE, HIDE_UNTIL_DUE_TIME));

    mergeDates(newTask(with(DUE_DATE, new DateTime(2016, 3, 11))).getDueDate(), local);

    assertEquals(
        new DateTime(2016, 3, 11, 13, 30, 1).getMillis(), local.getHideUntil().longValue());
  }

  @Test
  public void testDueTimeAdjustTimeForwards() {
    Task local =
        newTask(
            with(DUE_TIME, new DateTime(2016, 3, 12, 13, 30)),
            with(HIDE_TYPE, HIDE_UNTIL_DUE_TIME));

    mergeDates(newTask(with(DUE_DATE, new DateTime(2016, 3, 14))).getDueDate(), local);

    assertEquals(
        new DateTime(2016, 3, 14, 13, 30, 1).getMillis(), local.getHideUntil().longValue());
  }

  @Test
  public void testDueDateClearHide() {
    Task local =
        newTask(with(DUE_DATE, new DateTime(2016, 3, 12)), with(HIDE_TYPE, HIDE_UNTIL_DUE));

    mergeDates(newTask().getDueDate(), local);

    assertEquals(0, local.getHideUntil().longValue());
  }

  @Test
  public void testDueTimeClearHide() {
    Task local =
        newTask(
            with(DUE_TIME, new DateTime(2016, 3, 12, 13, 30)),
            with(HIDE_TYPE, HIDE_UNTIL_DUE_TIME));

    mergeDates(newTask().getDueDate(), local);

    assertEquals(0, local.getHideUntil().longValue());
  }

  @Test
  public void truncateValue() {
    assertEquals("1234567", truncate("12345678", 7));
  }

  @Test
  public void dontTruncateMax() {
    assertEquals("1234567", truncate("1234567", 7));
  }

  @Test
  public void dontTruncateShortValue() {
    assertEquals("12345", truncate("12345", 7));
  }

  @Test
  public void dontTruncateNull() {
    assertNull(truncate(null, 7));
  }

  @Test
  public void dontOverwriteTruncatedValue() {
    assertEquals("123456789", getTruncatedValue("123456789", "1234567", 7));
  }

  @Test
  public void overwriteTruncatedValueWithShortenedValue() {
    assertEquals("12345", getTruncatedValue("123456789", "12345", 7));
  }

  @Test
  public void overwriteTruncatedValueWithNullValue() {
    assertNull(getTruncatedValue("123456789", null, 7));
  }

  @Test
  public void overwriteNullValueWithTruncatedValue() {
    assertEquals("1234567", getTruncatedValue(null, "1234567", 7));
  }
}
